package eu.ruekov.ruakij.LinkedBeaconTeleporters.linkedBeaconTeleporter;

import eu.ruekov.ruakij.LinkedBeaconTeleporters.Function;
import eu.ruekov.ruakij.LinkedBeaconTeleporters.Main;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.NamespacedKey;
import org.bukkit.block.Block;
import org.bukkit.configuration.file.YamlConfiguration;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.ShapelessRecipe;
import org.bukkit.inventory.meta.ItemMeta;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class LinkedBeaconTeleporterManager {
    // Stores all placed Beacon-Teleporters for fast access
    // TODO: Evaluate necessity of 2 HashMaps with main reason of performance
    public static HashMap<String, List<LinkedBeaconTeleporter>> placedLBTsById = new HashMap();
    public static HashMap<String, LinkedBeaconTeleporterBlock> placedLBTByLoc = new HashMap();

    // Regex to match item-name&lore and extract id (group 1)
    static String regex;

    public static void init(){
        Main.log.info("Loading LinkedBeaconTeleporter from config..");

        // Construct search item-regex
        regex = constructRegex(
                Main.config.getString("item.name"),
                Main.config.getStringList("item.lore")
        );

        // Construct & add crafting-receipes
       constructAndAddRecipes();

        // Load teleporters
        for(String teleporterId : Main.data.getKeys(false)){
            List<String> serialisedLocs = Main.data.getStringList(teleporterId);

            // New list for this id
            List<LinkedBeaconTeleporter> linkedBeaconTeleporters = new ArrayList();

            for(String serializedLoc : serialisedLocs) {

                // Load location
                Location loc;
                try {
                    loc = Function.deserializeBlockLocation(serializedLoc);
                } catch (InvalidPropertiesFormatException e) {
                    Main.log.severe("Could not load location='"+ serializedLoc +"' from "+ teleporterId);
                    continue;
                }

                // Check if block is of type Beacon (in case the world was changed without the plugin running)
                Block block = loc.getBlock();
                if(block.getType() != Material.BEACON){
                    // Data out of sync! Skipping
                    Main.log.warning(teleporterId +" at location "+ serializedLoc +" changed and is no longer LinkedBlockTeleporter! (Changed without the plugin running?)");
                    continue;
                }

                // Construct block
                LinkedBeaconTeleporterBlock lbtBlock = new LinkedBeaconTeleporterBlock(teleporterId, block);
                // Write reference of list
                lbtBlock.linkedBeaconTeleporters = linkedBeaconTeleporters;

                // Write block to list
                linkedBeaconTeleporters.add(lbtBlock);
                // Save to loc-list
                placedLBTByLoc.put(serializedLoc, lbtBlock);
            }

            // Save list to id-list
            placedLBTsById.put(teleporterId, linkedBeaconTeleporters);
        }

        Main.log.info("All done!");
    }

    public static void saveData(){
        Main.log.info("Saving data..");

        Main.data = new YamlConfiguration();
        for(String teleporterId : placedLBTsById.keySet()){
            List<LinkedBeaconTeleporter> LBTs = placedLBTsById.get(teleporterId);

            // Generate clean array with data we want
            List<String> dataList = new ArrayList<>();
            for(LinkedBeaconTeleporter lbt : LBTs){
                dataList.add(
                        Function.serialiseBlockLocation(
                                ((LinkedBeaconTeleporterBlock)lbt).block().getLocation()
                        )
                );
            }

            Main.data.set(teleporterId, dataList);
        }

        Main.saveData();

        Main.log.info("done");
    }

    static void constructAndAddRecipes(){
        // Create result-item
        ItemStack item = getItemStackFromData("-");
        item.setAmount(2);

        // Create recipe
        ShapelessRecipe recipe = new ShapelessRecipe(
                new NamespacedKey(Main.plugin, "linked_beacon_teleporter")
                , item);
        recipe.addIngredient(2, Material.BEACON);
        Bukkit.addRecipe(recipe);

        // Empty recipe to reset beacon
        ShapelessRecipe recipe2 = new ShapelessRecipe(
                new NamespacedKey(Main.plugin, "linked_beacon_teleporter_reset"),
                new ItemStack(Material.BEACON)
                );
        recipe2.addIngredient(1, Material.BEACON);
        Bukkit.addRecipe(recipe2);
    }

    // Conversion-methods
    public static ItemStack getItemStackFromData(String teleporterId){
        ItemStack item = new ItemStack(Material.BEACON);
        ItemMeta itemMeta = item.getItemMeta();

        // Get config-values
        String name = Main.config.getString("item.name");
        List<String> lore = Main.config.getStringList("item.lore");

        // Replace parameters
        name = name.replace("%id%", teleporterId);
        for(int i=0; i<lore.size(); i++){
            lore.set(i,
                    lore.get(i).replace("%id%", teleporterId)
            );
        }

        // Set meta
        itemMeta.setDisplayName(name);
        itemMeta.setLore(lore);

        item.setItemMeta(itemMeta);

        return item;
    }
    public static LinkedBeaconTeleporterItem getLbtItemFromItemStack(ItemStack itemStack){
        ItemMeta itemMeta = itemStack.getItemMeta();

        // Serialize
        String nameAndLore = serialiseBeaconItem(
                itemMeta.getDisplayName(),
                itemMeta.getLore()
        );

        // Regex match
        Pattern pattern = Pattern.compile(regex);
        Matcher matcher = pattern.matcher(nameAndLore);
        if(!matcher.find())
            // No data
            return null;

        LinkedBeaconTeleporterItem lbtItem = new LinkedBeaconTeleporterItem(
                matcher.group(1),
                itemStack
        );

        return lbtItem;
    }

    public static LinkedBeaconTeleporterBlock getLbtBlockFromLocation(Location loc){
        // Serialize
        String serializedLoc = Function.serialiseBlockLocation(loc);

        return placedLBTByLoc.get(serializedLoc);
    }
    public static List<LinkedBeaconTeleporter> getLbtBlocksFromTeleportId(String teleportId){
        return placedLBTsById.get(teleportId);
    }

    static String constructRegex(String name, List<String> lore){
        // Serialize
        String regex = serialiseBeaconItem(name, lore);

        // Wrap in regex-quotes
        regex = "^\\Q"+ regex +"\\E$";
        // Exclude id from regex-quotes and create group
        regex = regex.replace("%id%", "\\E(.*)\\Q");

        // Remove empty regex-quotes
        regex = regex.replace("\\E\\Q", "");

        return regex;
    }
    static String serialiseBeaconItem(String name, List<String> lore){
        // get name and remove control-chars
        String str = name.replace("\\", "\\\\");
        if(lore != null){
            for (String loreLine : lore) {
                // get lore-lines and remove control-chars
                str += "\n"+ loreLine.replace("\\", "\\\\");
            }
        }

        return str;
    }
}
